#!/usr/bin/env/python2
# -*- coding: utf-8 -*-

# Copyright (c) 2019, Piotr Dobrowolski <informatic@hackerspace.pl
# Based on IdeaBank crawler by Remigiusz Marcinkiewicz <remigiusz@marcinkiewicz.me>
# Based on iBRE/mBank CompanyNet crawler by Sergiusz Bazanski <q3k@q3k.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import date, datetime, timedelta
import sys
import time
import os
import re
import logging
import logging.config
import argparse
import json

import requests
from sqlalchemy import Column, Integer, String, Date, BigInteger, create_engine, MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import CreateTable, CreateIndex
from six.moves import input

from config import CurrentConfig


config = {
    key: getattr(CurrentConfig, key)
    for key in dir(CurrentConfig) if key.isupper()}

Base = declarative_base()

if config.get('LOGGING'):
    logging.config.dictConfig(config['LOGGING'])
else:
    logging.basicConfig(level=logging.DEBUG)

class RawTransfer(Base):
    __tablename__ = 'raw_transfer'
    id = Column(Integer, primary_key=True)
    raw = Column(String)
    uid = Column(String(128), index=True)
    on_account = Column(String(32), index=True)
    amount = Column(Integer)
    currency = Column(String(8))
    date = Column(Date)
    type = Column(String(16))
    index = Column(Integer)

    title = Column(String(256))

    balance = Column(Integer)
    balance_currency = Column(String(8))

    from_account = Column(String(32))
    to_account = Column(String(32))

    from_name = Column(String(256))
    to_name = Column(String(256))

    scrape_timestamp = Column(BigInteger, default=lambda: round(time.time() * 1000000))

class IBParseError(Exception):
    pass

class IBMaintenanceError(Exception):
    pass

class IBRow(RawTransfer):
    SECRET = config["SECRET"]
    OWN_ACCOUNTS = config["OWN_ACCOUNTS"]

    def __unicode__(self):
        if self.balance is not None:
            return u"{} *{} #{} @{} -\"{}\" -#{} => +\"{}\" +#{} [{}.{:02d} {}] ({}.{:02d} {}) ~\"{}\"".format(
                self.type, self.index, self.on_account, self.date, self.from_name, self.from_account,
                self.to_name, self.to_account, self.amount/100, self.amount%100,
                self.currency, self.balance/100, self.balance%100, self.balance_currency, self.title)
        return u"{} *{} #{} @{} -\"{}\" -#{} => +\"{}\" +#{} [{}.{:02d} {}] (?) ~\"{}\"".format(
            self.type, self.index, self.on_account, self.date, self.from_name, self.from_account,
            self.to_name, self.to_account, self.amount/100, self.amount%100,
            self.currency, self.title)
    def __str__(self):
        return unicode(self).encode("utf-8")
    def __repr__(self):
        return str(self)

    def __init__(self, row, own_account):
        self.date = datetime.strptime(row['date'], '%d.%m.%Y').date()
        self.index = 1
        self.raw = json.dumps(row, separators=(',', ':'))
        self.uid = row['id']
        self.title = row['title']

        self.amount = int(round(row['amount'] * 100))
        self.currency = row['currency']

        self.on_account = own_account

        self.from_name = row['remitterName']
        self.from_account = IBParser.parse_account_number(row['remitterNrb'])
        self.to_name = row['beneficiaryName']
        self.to_account = IBParser.parse_account_number(row['beneficiaryNrb'])

        direction = row['kind']
        if direction == 'OUT' or direction == 'CARD_TRANS':
            self.type = 'OUT'
            self.amount = -self.amount
            if row['operationType'] == 'SELF':
                self.type = 'OUT_TO_OWN'
        elif direction == 'IN':
            self.type = 'IN'
            if row['operationType'] == 'SELF':
                self.type = 'IN_FROM_OWN'
        elif direction == 'FEE':
            self.type = 'BANK_FEE'
            self.amount = -self.amount
            if self.to_name != self.from_name and 'Prowizja za przelew natychmiastowy' not in self.title:
                # TODO FIXME: false for instant transfer fees
                raise IBParseError("Invalid to_name/from_name (%r / %r)" % (
                    self.to_name, self.from_name))
            if self.from_account not in self.OWN_ACCOUNTS and self.to_account not in self.OWN_ACCOUNTS:
            # if self.from_account not in self.OWN_ACCOUNTS or self.to_account not in self.OWN_ACCOUNTS:
                raise IBParseError("Wrong to_account/from_account on bank fee transfer (%r / %r)" % (
                    self.to_account, self.from_account
                ))
            # TODO FIXME: false for instant transfer fees
            # To account seems to always be main account
            # self.to_account = self.from_account
        else:
            raise IBParseError(
                "Can't parse direction specifier \"{}\"".format(direction))

        if None in (self.type, self.to_account, self.from_account, self.to_name, self.from_name):
            print(row)
            raise IBParseError(
                "Something went wrong - one of the mandatory values empty",
                self.type, self.to_account, self.from_account,
                self.to_name, self.from_name)

class IBParser(object):
    def __init__(self, account_number):
        self.account_number = account_number
        self.rows = []
        self.fields = []

    def parse(self, snapshot):
        for tx in snapshot:
            self.rows.append(IBRow(tx, self.account_number))

    def get(self, type_=None, on_account=None):
        return [
            row for row in self.rows
            if (row.type == type_ or type_ is None)
            and (row.on_account == on_account or on_account is None)]

    @staticmethod
    def parse_account_number(s):
        formats = [
            "((?:[A-Za-z]{2})?[0-9]{2})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})", # 26 digits, optional country code - Poland
            "((?:[A-Za-z]{2})?[0-9]{2})[ ]?([A-Z]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{2})", # 22 characters including BIC bank code - Ireland
            "((?:[A-Za-z]{2})?[0-9]{2})[ ]?([A-Z]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{2})", # 18 characters including BIC bank code - Netherlands
            "((?:[A-Za-z]{2})?[0-9]{14})", # 14 characters including BIC bank code - Belgium
            "((?:[A-Za-z]{2})?[0-9]{2})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{4})[ ]?([0-9]{2})", # 22 digits - Germany
            "^([0-9]{5})$", # 5 digits - weird special hax for Benevity (UK?)
            ]
        for f in formats:
            m = re.search(f, s)
            if m is not None:
                break

        if m is None:
            return None

        account = "".join(m.groups())
        if len(m.group(1)) == 2:
            account = "PL" + account

        return account

class IBFetcher(object):
    BASE = "https://secure.ideabank.pl/"
    START_DATE = "01.11.2016"
    def __init__(self, config):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.token = None
        self.config = config
        self.s = requests.Session()
        self.s.headers.update({
            "User-Agent": config.get("IB_UA", 'Mozilla/5.0 (X11; Linux x86_64; rv:69.0) Gecko/20100101 Firefox/69.0'),
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate",
            "DNT": "1",
            "Upgrade-Insecure-Requests": "1",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache"
            })

        self.s.cookies.update({
            'ib_trusted_device': config.get('IB_TRUSTED_DEVICE_TOKEN', ''),
        })

    def _wait(self, seconds):
        self.logger.debug("Waiting %d seconds", seconds)
        time.sleep(seconds)

    def _request(self, url, method='GET', *args, **kwargs):
        if not url.startswith('http'):
            url = 'https://cloud.ideabank.pl/api' + url

        if 'params' not in kwargs:
            kwargs['params'] = {}

        kwargs['params'].update({
            'v': round(time.time() * 1000),
        })

        if self.token:
            if 'headers' not in kwargs:
                kwargs['headers'] = {}

            kwargs['headers'].update({
                'Authentication-Token': self.token,
            })

        resp = self.s.request(method, url, *args, **kwargs)
        resp.raise_for_status()

        return resp

    def _get(self, url, *args, **kwargs):
        return self._request(url, 'GET', *args, **kwargs)

    def _post(self, url, *args, **kwargs):
        return self._request(url, 'POST', *args, **kwargs)

    def login(self, login, password):
        self._get('https://sso.cloud.ideabank.pl/authenticate/login', params={
            'login': login,
        })

        login_resp = self._post('https://sso.cloud.ideabank.pl/authenticate/login', json={
            'login': login,
            'password': password,
        }).json()

        login_token = login_resp['token']

        apilogin_resp = self._post('/login', data={
            'token': login_token,
        })
        auth_token = apilogin_resp.history[0].cookies['Authentication-Token']
        self.token = auth_token

        self.logger.debug('Authentication token: %s', self.token)

        return auth_token

    def get_wallet(self):
        accounts = self._get('/accounts').json()
        wallet = {
            'accounts': {},
        }
        for g in accounts['firmAccountGroups']:
            for a in g['accounts']:
                account_number = IBParser.parse_account_number(a['nrb']['value'])
                wallet['accounts'][account_number] = {
                    'id': a['productId'],
                    'available_balance': a['balance'],  # FIXME activeBalance?
                    'currency': a['currencyCode'],
                }

        return wallet

    def fetch_account_history(self, account_id, start=None, end=None):
        epoch_start = date(2019, 10, 20)
        if end is None:
            end = date.today()
        if start is None:
            start = date.today() - timedelta(days=30)
        if start < epoch_start:
            start = epoch_start

        total_pages = 1
        page = 0
        history = []
        while page < total_pages:
            transactions = self._get('/transactions', params={
                'from': '{:04d}-{:02d}-{:02d}'.format(start.year, start.month, start.day),
                'to': '{:04d}-{:02d}-{:02d}'.format(end.year, end.month, end.day),
                'page': str(page),
                'productIds': account_id,
            }).json()
            history.extend(transactions['history'])

            page += 1
            total_pages = transactions['page']['totalPages']

        return history


def usage():
    pass


def lock():
    fn = config["LOCK_FILE"]
    if os.path.isfile(fn):
        logging.error("Lock file %s exists, aborting", fn)
        sys.exit(3)
    logging.debug("Setting up lock file %s", fn)
    open(fn,'w').close()
    if not os.path.isfile(fn):
        logging.error("Lock file %s somehow does not exist, aborting", fn)
        sys.exit(3)

def release():
    fn = config["LOCK_FILE"]
    logging.debug("Removing lock file %s", fn)
    if not os.path.isfile(fn):
        logging.error("Lock file %s somehow does not exist, WTF?", fn)
        sys.exit(3)
    os.remove(fn)
    if os.path.isfile(fn):
        logging.error("Lock file %s somehow still exists, WTF?", fn)
        sys.exit(3)

parser = argparse.ArgumentParser()
parser.add_argument('-n', '--no-action', action="store_true", help='do not commit any database changes')
parser.add_argument('-c', '--cached', action="store_true", help='use cached data (test)')
parser.add_argument('-l', '--load', action='append', help='process specified files (test)')
parser.add_argument('-t', '--token', help='use authentication token')
parser.add_argument('--start', type=lambda s: datetime.strptime(s, '%Y-%m-%d').date(), help='start date (YYYY-MM-DD)')
parser.add_argument('--end', type=lambda s: datetime.strptime(s, '%Y-%m-%d').date(), help='end date (YYYY-MM-DD)')
parser.add_argument('--no-lock', action='store_true', help='don\'t use lockfile (test)')
parser.add_argument('--print-schema', action="store_true", help='print table schema and quit')

if __name__ == "__main__":
    args = parser.parse_args()

    CACHE_DIR = config["CACHE_DIR"]
    engine = create_engine(config["SQLALCHEMY_DATABASE_URI"])
    session = sessionmaker(bind=engine)()

    if args.print_schema:
        logging.debug("Called with --print-schema, will print the create " +
                      "statement and quit.")
        m = MetaData()
        print('%s;' % CreateTable(IBRow.__table__).compile(engine))
        for index in IBRow.__table__.indexes:
            print('%s;' % CreateIndex(index).compile(engine))
        sys.exit()

    if not args.no_lock:
        lock()

    balances = {}
    history_logs = {}

    if args.load:
        logging.debug("Using manually supplied files")
        for fn in args.load:
            an, f = fn.split(':')
            account_number = IBParser.parse_account_number(an)
            if account_number is None:
                logging.error("File name number \"{}\" unparseable".format(f))
                continue

            logging.debug('Loading "%s" as "%s"', f, account_number)

            with open(f, 'r') as fd:
                history_logs[account_number] = json.loads(fd.read())
    elif args.cached:
        logging.debug("Loading cached files from {}".format(CACHE_DIR))
        for f in os.listdir(CACHE_DIR):
            if f.startswith('balance-'):
                continue

            account_number = IBParser.parse_account_number(f)
            if account_number is None:
                logging.error("File name number \"{}\" unparseable".format(f))
                continue

            with open(CACHE_DIR + "/" + f, 'r') as fd:
                try:
                    history_logs[account_number] = json.loads(fd.read())
                except Exception as e:
                    logging.error("Failed to decode {}: {}".format(f, e))

            logging.debug("Loading \"{}\" as \"{}\"".format(f, account_number))
    else:
        logging.debug("Normal run - will connect to the bank")
        fetcher = IBFetcher(config)
        if args.token:
            fetcher.token = args.token
            logging.debug("Using provided token")
        elif "IB_LOGIN" not in config.keys() or "IB_PASSWORD" not in config.keys():
            fetcher.login(input("[?] ID: "), input("[?] Password: "))
        else:
            logging.debug("Using saved credentials")
            fetcher.login(config["IB_LOGIN"], config["IB_PASSWORD"])

        wallet = fetcher.get_wallet()
        for account_number, account in wallet["accounts"].items():
            logging.debug("Fetching history for account {} ({})".format(
                account_number, account["id"]))
            history = fetcher.fetch_account_history(account["id"], start=args.start, end=args.end)

            history_logs[account_number] = history
            with open(CACHE_DIR + "/" + account_number, 'w') as fd:
                fd.write(json.dumps(history))

            balances[account_number] = (
                account["available_balance"], account["currency"])
            with open(CACHE_DIR + "/balance-"+account_number, 'w') as fd:
                fd.write("{} {}\n".format(
                    account["available_balance"], account["currency"]))

    if not history_logs:
        logging.error('Nothing to process')
        sys.exit()

    parsed = {}
    stats = {}
    for account_number, history in history_logs.items():
        logging.debug("Parsing history for account {}".format(account_number))
        parser = IBParser(account_number)
        try:
            parser.parse(history)
        except IBMaintenanceError:
            logging.exception('Maintenance error, skipping')
            continue

        stats[account_number] = {}
        stats[account_number]["added"] = 0
        stats[account_number]["skipped"] = 0
        for row in parser.get():
            if not session.query(IBRow).filter_by(uid=row.uid).first():
                if args.no_action:
                    print(row)
                session.add(row)
                stats[account_number]["added"] += 1
            else:
                stats[account_number]["skipped"] += 1

    if args.no_action:
        logging.info('Running with --no-action, not commiting.')
    else:
        session.commit()

    # That is pretty ugly, but the only alternative would be to change handler
    # level in runtime, and that'd still need some rollback anyway.
    if any(v['added'] for v in stats.values()):
        log_summary = logging.info
    else:
        log_summary = logging.debug

    if balances:
        log_summary("Account balances:")
        for account_number, v in balances.items():
            balance, currency = v
            log_summary("\t{}: {} {}".format(
                account_number, balance, currency))

    log_summary("Done: %r", stats)
    try:
        if stats.get('PL91195000012006000648890004',{}).get('added'):
            msg = 'holla holla get dolla: {1[0]} {1[1]} (+{0})'.format(stats.get('PL91195000012006000648890004',{}).get('added'), balances.get('PL91195000012006000648890004'))
            requests.post('http://hackerspace.pl:43288/moonspeak/1/notification', params={
                'target': '#hackerspace-pl-members', 'message': msg,
            })
    except Exception as exc:
        print(exc)

    if not args.no_lock:
        release()
