# Copyright (c) 2015, Sergiusz Bazanski <q3k@q3k.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from functools import wraps

import memcache
import requests
import sqltap.wsgi
import click
import json
import decimal
import datetime
from flask import Flask, redirect
from flask_sqlalchemy import SQLAlchemy
from flask_login import LoginManager, AnonymousUserMixin, login_required, current_user
from flask_caching import Cache
from flask_gravatar import Gravatar
from spaceauth import SpaceAuth

app = Flask(__name__)
app.config.from_object("config.CurrentConfig")

auth = SpaceAuth()
db = SQLAlchemy()
cache = Cache()
gravatar = Gravatar(size=256, rating='g', default='retro',
                    force_default=False, use_ssl=True, base_url=None)

# TODO unsubscribe me from life
cache_enabled = False
mc = cache


import webapp.models  # noqa
from webapp.authutils import AnonymousUser, User

auth.anonymous_user = AnonymousUser
@auth.user_loader
def load_user(username, profile=None):
    return User(username)

def admin_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not current_user.is_admin:
            return auth.unauthorized()
        return func(*args, **kwargs)
    return wrapper


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            # wanted a simple yield str(o) in the next line,
            # but that would mean a yield on the line with super(...),
            # which wouldn't work (see my comment below), so...
            return str(o)
            #return (str(o) for o in [o])
        elif isinstance(o, datetime.datetime):
            return str(o)
            #return (str(o) for o in [o])
        return super(DecimalEncoder, self).default(o)


def create_app():
    db.init_app(app)
    auth.init_app(app)
    gravatar.init_app(app)
    cache.init_app(app)

    # Initialize middleware
    if app.debug:
        app.wsgi_app = sqltap.wsgi.SQLTapMiddleware(app.wsgi_app)

    # Setup prometheus metrics
    if app.config.get('PROMETHEUS_DIR'):
        # This needs to be set before importing prometheus_client
        os.environ['prometheus_multiproc_dir'] = app.config['PROMETHEUS_DIR']

        # FIXME: we could expose this somehow
        from prometheus_flask_exporter.multiprocess import UWsgiPrometheusMetrics
        metrics = UWsgiPrometheusMetrics(group_by='url_rule')

        metrics.init_app(app)
        #metrics.register_endpoint('/varz', app)

    # Register blueprints
    import webapp.views
    import webapp.admin
    import webapp.api
    app.register_blueprint(webapp.admin.bp)
    app.register_blueprint(webapp.api.bp)

    # Custom filters
    @app.template_filter('inflect')
    def inflect(v, one, two, five):
        num = abs(v)

        if num == 0:
            return '%d %s' % (v, five)
        elif num == 1:
            return '%d %s' % (v, one)
        elif num <= 4:
            return '%d %s' % (v, two)

        return '%d %s' % (v, five)

    # Custom CLI commands
    import webapp.commands
    app.cli.commands.update(webapp.commands.group.commands)

    app.json_encoder = DecimalEncoder

    return app
